<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Bootstrap;

// Import framework stuff
use Org\Mxchange\CoreFramework\Configuration\FrameworkConfiguration;
use Org\Mxchange\CoreFramework\Connection\Database\DatabaseConnection;
use Org\Mxchange\CoreFramework\Connector\Database\DatabaseConnector;
use Org\Mxchange\CoreFramework\Console\Tools\ConsoleTools;
use Org\Mxchange\CoreFramework\EntryPoint\ApplicationEntryPoint;
use Org\Mxchange\CoreFramework\Factory\ObjectFactory;
use Org\Mxchange\CoreFramework\Generic\NullPointerException;
use Org\Mxchange\CoreFramework\Helper\Application\ApplicationHelper;
use Org\Mxchange\CoreFramework\Localization\ManageableLanguage;
use Org\Mxchange\CoreFramework\Loader\ClassLoader;
use Org\Mxchange\CoreFramework\Manager\ManageableApplication;
use Org\Mxchange\CoreFramework\Middleware\Debug\DebugMiddleware;
use Org\Mxchange\CoreFramework\Object\BaseFrameworkSystem;
use Org\Mxchange\CoreFramework\Registry\GenericRegistry;
use Org\Mxchange\CoreFramework\Request\Requestable;
use Org\Mxchange\CoreFramework\Response\Responseable;
use Org\Mxchange\CoreFramework\Utils\String\StringUtils;

// Import SPL stuff
use \BadMethodCallException;
use \InvalidArgumentException;
use \SplFileInfo;

/**
 * A framework-bootstrap class which helps the frameworks to bootstrap ... ;-)
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2020 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
final class FrameworkBootstrap {

	/**
	 * Detected server address
	 */
	private static $serverAddress = NULL;

	/**
	 * Instance of a Requestable class
	 */
	private static $requestInstance = NULL;

	/**
	 * Instance of a Responseable class
	 */
	private static $responseInstance = NULL;

	/**
	 * Instance of a FrameworkConfiguration class
	 */
	private static $configurationInstance = NULL;

	/**
	 * Database instance
	 */
	private static $databaseInstance = NULL;

	/**
	 * Language system instance
	 */
	private static $languageInstance = NULL;

	/*
	 * Includes applications may have. They will be tried in the given order,
	 * some will become soon deprecated.
	 */
	private static $configAppIncludes = array(
		// The ApplicationHelper class (required)
		'class_ApplicationHelper' => 'required',
		// Some debugging stuff (optional but can be committed)
		'debug'                   => 'optional',
		// Application's exception handler (optional but can be committed)
		'exceptions'              => 'optional',
		// Application's configuration file (committed, non-local specific)
		'config'                  => 'required',
		// Local configuration file (optional, not committed, listed in .gitignore)
		'config-local'            => 'optional',
		// Application data (deprecated)
		'data'                    => 'deprecated',
		// Application loader (deprecated)
		'loader'                  => 'deprecated',
		// Application initializer (deprecated)
		'init'                    => 'deprecated',
		// Application starter (deprecated)
		'starter'                 => 'deprecated',
	);

	/**
	 * Private constructor, no instance is needed from this class as only
	 * static methods exist.
	 */
	private function __construct () {
		// Prevent making instances from this "utilities" class
	}

	/**
	 * Some "getter" for a configuration instance, making sure, it is unique
	 *
	 * @return	$configurationInstance	An instance of a FrameworkConfiguration class
	 */
	public static function getConfigurationInstance () {
		// Is the instance there?
		if (is_null(self::$configurationInstance)) {
			// Init new instance
			self::$configurationInstance = new FrameworkConfiguration();
		} // END - if

		// Return it
		return self::$configurationInstance;
	}

	/**
	 * "Getter" to get response/request type from analysis of the system.
	 *
	 * @return	$requestType	Analyzed request type
	 */
	public static function getRequestTypeFromSystem () {
		// Default is console
		$requestType = 'console';

		// Is 'HTTP_HOST' set?
		if (isset($_SERVER['HTTP_HOST'])) {
			// Then it is a HTML response/request.
			$requestType = 'html';
		} // END - if

		// Return it
		return $requestType;
	}

	/**
	 * Checks whether the given file/path is in open_basedir(). This does not
	 * gurantee that the file is actually readable and/or writeable. If you need
	 * such gurantee then please use isReadableFile() instead.
	 *
	 * @param	$fileInstance	An instance of a SplFileInfo class
	 * @return	$isReachable	Whether it is within open_basedir()
	 */
	public static function isReachableFilePath (SplFileInfo $fileInstance) {
		// Is not reachable by default
		$isReachable = false;

		// Get open_basedir parameter
		$openBaseDir = trim(ini_get('open_basedir'));

		// Is it set?
		if (!empty($openBaseDir)) {
			// Check all entries
			foreach (explode(PATH_SEPARATOR, $openBaseDir) as $dir) {
				// Check on existence
				if (substr($fileInstance->getPathname(), 0, strlen($dir)) == $dir) {
					// Is reachable
					$isReachable = true;

					// Abort lookup as it has been found in open_basedir
					break;
				} // END - if
			} // END - foreach
		} else {
			// If open_basedir is not set, all is allowed
			$isReachable = true;
		}

		// Return status
		return $isReachable;
	}

	/**
	 * Checks whether the give file is within open_basedir() (done by
	 * isReachableFilePath()), is actually a file and is readable.
	 *
	 * @param	$fileInstance	An instance of a SplFileInfo class
	 * @return	$isReadable		Whether the file is readable (and therefor exists)
	 */
	public static function isReadableFile (SplFileInfo $fileInstance) {
		// Default is not readable
		$isReadable = false;

		// Check if it is a file and readable
		$isReadable = (
			(
				self::isReachableFilePath($fileInstance)
			) && (
				$fileInstance->isFile()
			) && (
				$fileInstance->isReadable()
			)
		);

		// Return status
		return $isReadable;
	}

	/**
	 * Loads given include file
	 *
	 * @param	$fileInstance	An instance of a SplFileInfo class
	 * @return	void
	 * @throws	InvalidArgumentException	If file was not found or not readable or deprecated
	 */
	public static function loadInclude (SplFileInfo $fileInstance) {
		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: fileInstance=%s - CALLED!' . PHP_EOL, __METHOD__, __LINE__, $fileInstance);

		// Should be there ...
		if (!self::isReadableFile($fileInstance)) {
			// Abort here
			throw new InvalidArgumentException(sprintf('Cannot find fileInstance.pathname=%s.', $fileInstance->getPathname()));
		} // END - if

		// Load it
		require_once $fileInstance->getPathname();

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Does the actual bootstrap
	 *
	 * @return	void
	 */
	public static function doBootstrap () {
		// Load basic include files to continue bootstrapping
		self::loadInclude(new SplFileInfo(sprintf('%smain%sinterfaces%sclass_FrameworkInterface.php', ApplicationEntryPoint::detectFrameworkPath(), DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR)));
		self::loadInclude(new SplFileInfo(sprintf('%smain%sclasses%sclass_BaseFrameworkSystem.php', ApplicationEntryPoint::detectFrameworkPath(), DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR)));
		self::loadInclude(new SplFileInfo(sprintf('%smain%sclasses%sutils%sstring%sclass_StringUtils.php', ApplicationEntryPoint::detectFrameworkPath(), DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR)));
		self::loadInclude(new SplFileInfo(sprintf('%smain%sinterfaces%sregistry%sclass_Registerable.php', ApplicationEntryPoint::detectFrameworkPath(), DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR)));
		self::loadInclude(new SplFileInfo(sprintf('%sconfig%sclass_FrameworkConfiguration.php', ApplicationEntryPoint::detectFrameworkPath(), DIRECTORY_SEPARATOR)));

		// Load global configuration
		self::loadInclude(new SplFileInfo(sprintf('%s%s', ApplicationEntryPoint::detectFrameworkPath(), 'config-global.php')));
	}

	/**
	 * Initializes the framework by scanning for all framework-relevant
	 * classes, interfaces and exception. Then determine the request type and
	 * initialize a Requestable instance which will then contain all request
	 * parameter, also from CLI. Next step is to validate the application
	 * (very basic).
	 *
	 * @return	void
	 */
	public static function initFramework () {
		/**
		 * 1) Load class loader and scan framework classes, interfaces and
		 *    exceptions.
		 */
		self::scanFrameworkClasses();

		/*
		 * 2) Determine the request type, console or web and store request and
		 *    response here. This also initializes the request instance will
		 *    all given parameters (see doc-tag for possible sources of
		 *    parameters).
		 */
		self::determineRequestType();

		/*
		 * 3) Now, that there are all request parameters being available, check
		 *    if 'application' is supplied. If it is not found, abort execution, if
		 *    found, continue below with next step.
		 */
		self::validateApplicationParameter();
	}

	/**
	 * Initializes the detected application. This may fail if required files
	 * are not found in the application's base path (not to be confused with
	 * 'application_base_path' which only points to /some/foo/application/.
	 *
	 * @return	void
	 */
	public static function prepareApplication () {
		// Configuration entry 'detected_app_name' must be set, get it here, including full path
		$application = self::getConfigurationInstance()->getConfigEntry('detected_app_name');
		$fullPath    = self::getConfigurationInstance()->getConfigEntry('detected_full_app_path');

		/*
		 * Now check and load all files, found deprecated files will throw a
		 * warning at the user.
		 */
		foreach (self::$configAppIncludes as $fileName => $status) {
			// Construct file instance
			$fileInstance = new SplFileInfo(sprintf('%s%s.php', $fullPath, $fileName));

			// Determine if this file is wanted/readable/deprecated
			if (($status == 'required') && (!self::isReadableFile($fileInstance))) {
				// Nope, required file cannot be found/read from
				ApplicationEntryPoint::exitApplication(sprintf('Application "%s" does not have required file "%s.php". Please add it.', $application, $fileInstance->getBasename()));
			} elseif (($fileInstance->isFile()) && (!$fileInstance->isReadable())) {
				// Found, not readable file
				ApplicationEntryPoint::exitApplication(sprintf('File "%s.php" from application "%s" cannot be read. Please fix CHMOD.', $fileInstance->getBasename(), $application));
			} elseif (($status != 'required') && (!self::isReadableFile($fileInstance))) {
				// Not found but optional/deprecated file, skip it
				continue;
			}

			// Is the file deprecated?
			if ($status == 'deprecated') {
				// Issue warning
				trigger_error(sprintf('Deprecated file "%s.php" found, will not load it to avoid problems. Please remove it from your application "%s" to avoid this warning.', $fileName, $application), E_USER_WARNING);

				// Skip loading deprecated file
				continue;
			} // END - if

			// Load it
			self::loadInclude($fileInstance);
		} // END - foreach

		// Scan for application's classes, exceptions and interfaces
		ClassLoader::scanApplicationClasses();
	}

	/**
	 * Starts a fully initialized application, the class ApplicationHelper must
	 * be loaded at this point.
	 *
	 * @return	void
	 */
	public static function startApplication () {
		// Configuration entry 'detected_app_name' must be set, get it here
		$application = self::getConfigurationInstance()->getConfigEntry('detected_app_name');

		// Is there an application helper instance?
		$applicationInstance = call_user_func_array(
			array(
				'Org\Mxchange\CoreFramework\Helper\Application\ApplicationHelper', 'getSelfInstance'
			), []
		);

		// Some sanity checks
		if ((empty($applicationInstance)) || (is_null($applicationInstance))) {
			// Something went wrong!
			ApplicationEntryPoint::exitApplication(sprintf('[Main:] The application <span class="app_name">%s</span> could not be launched because the helper class <span class="class_name">%s</span> is not loaded.',
				$application,
				'Org\Mxchange\CoreFramework\Helper\Application\ApplicationHelper'
			));
		} elseif (!is_object($applicationInstance)) {
			// No object!
			ApplicationEntryPoint::exitApplication(sprintf('[Main:] The application <span class="app_name">%s</span> could not be launched because &#39;app&#39; is not an object (%s).',
				$application,
				gettype($applicationInstance)
			));
		} elseif (!($applicationInstance instanceof ManageableApplication)) {
			// Missing interface
			ApplicationEntryPoint::exitApplication(sprintf('[Main:] The application <span class="app_name">%s</span> could not be launched because &#39;app&#39; is lacking required interface ManageableApplication.',
				$application
			));
		}

		// Now call all methods in one go
		foreach (array('setupApplicationData', 'initApplication', 'launchApplication') as $methodName) {
			// Debug message
			//*NOISY-DEBUG: */ printf('[%s:%d]: Calling methodName=%s ...' . PHP_EOL, __METHOD__, __LINE__, $methodName);

			// Call method
			call_user_func(array($applicationInstance, $methodName));
		} // END - foreach
	}

	/**
	 * Initializes database instance, no need to double-call this method
	 *
	 * @return	void
	 */
	public static function initDatabaseInstance () {
		// Get application instance
		$applicationInstance = ApplicationHelper::getSelfInstance();

		// Is the database instance already set?
		if (self::getDatabaseInstance() instanceof DatabaseConnector) {
			// Yes, then abort here
			throw new BadMethodCallException('Method called twice.');
		} // END - if

		// Initialize database layer
		$databaseInstance = ObjectFactory::createObjectByConfiguredName(self::getConfigurationInstance()->getConfigEntry('database_type') . '_class');

		// Prepare database instance
		$connectionInstance = DatabaseConnection::createDatabaseConnection(DebugMiddleware::getSelfInstance(), $databaseInstance);

		// Set it in application helper
		self::setDatabaseInstance($connectionInstance);
	}

	/**
	 * Detects the server address (SERVER_ADDR) and set it in configuration
	 *
	 * @return	$serverAddress	The detected server address
	 * @throws	UnknownHostnameException	If SERVER_NAME cannot be resolved to an IP address
	 * @todo	Have to check some more entries from $_SERVER here
	 */
	public static function detectServerAddress () {
		// Is the entry set?
		if (!isset(self::$serverAddress)) {
			// Is it set in $_SERVER?
			if (!empty($_SERVER['SERVER_ADDR'])) {
				// Set it from $_SERVER
				self::$serverAddress = $_SERVER['SERVER_ADDR'];
			} elseif (isset($_SERVER['SERVER_NAME'])) {
				// Resolve IP address
				$serverIp = ConsoleTools::resolveIpAddress($_SERVER['SERVER_NAME']);

				// Is it valid?
				if ($serverIp === false) {
					/*
					 * Why is gethostbyname() returning the host name and not
					 * false as many other PHP functions are doing? ;-(
					 */
					throw new UnknownHostnameException(sprintf('Cannot resolve "%s" to an IP address. Please fix your setup.', $_SERVER['SERVER_NAME']));
				} // END - if

				// Al fine, set it
				self::$serverAddress = $serverIp;
			} else {
				// Run auto-detecting through console tools lib
				self::$serverAddress = ConsoleTools::acquireSelfIpAddress();
			}
		} // END - if

		// Return it
		return self::$serverAddress;
	}

	/**
	 * Setter for default time zone (must be correct!)
	 *
	 * @param	$timezone	The timezone string (e.g. Europe/Berlin)
	 * @return	$success	If timezone was accepted
	 * @throws	InvalidArgumentException	If $timezone is empty
	 */
	public static function setDefaultTimezone (string $timezone) {
		// Is it set?
		if (empty($timezone)) {
			// Entry is empty
			throw new InvalidArgumentException('timezone is empty');
		}

		// Default success
		$success = FALSE;

		/*
		 * Set desired time zone to prevent date() and related functions to
		 * issue an E_WARNING.
		 */
		$success = date_default_timezone_set($timezone);

		// Return status
		return $success;
	}

	/**
	 * Checks whether HTTPS is set in $_SERVER
	 *
	 * @return	$isset	Whether HTTPS is set
	 * @todo	Test more fields
	 */
	public static function isHttpSecured () {
		return (
			(
				(
					isset($_SERVER['HTTPS'])
				) && (
					strtolower($_SERVER['HTTPS']) == 'on'
				)
			) || (
				(
					isset($_SERVER['HTTP_X_FORWARDED_PROTO'])
				) && (
					strtolower($_SERVER['HTTP_X_FORWARDED_PROTO']) == 'https'
				)
			)
		);
	}

	/**
	 * Dectect and return the base URL for all URLs and forms
	 *
	 * @return	$baseUrl	Detected base URL
	 */
	public static function detectBaseUrl () {
		// Initialize the URL
		$protocol = 'http';

		// Do we have HTTPS?
		if (self::isHttpSecured()) {
			// Add the >s< for HTTPS
			$protocol = 'https';
		} // END - if

		// Construct the full URL and secure it against CSRF attacks
		$baseUrl = sprintf('%s://%s%s', $protocol, self::detectDomain(), self::detectScriptPath());

		// Return the URL
		return $baseUrl;
	}

	/**
	 * Detect safely and return the full domain where this script is installed
	 *
	 * @return	$fullDomain		The detected full domain
	 */
	public static function detectDomain () {
		// Full domain is localnet.invalid by default
		$fullDomain = 'localnet.invalid';

		// Is the server name there?
		if (isset($_SERVER['SERVER_NAME'])) {
			// Detect the full domain
			$fullDomain = htmlentities(strip_tags($_SERVER['SERVER_NAME']), ENT_QUOTES);
		} // END - if

		// Return it
		return $fullDomain;
	}

	/**
	 * Detect safely the script path without trailing slash which is the glue
	 * between "http://your-domain.invalid/" and "script-name.php"
	 *
	 * @return	$scriptPath		The script path extracted from $_SERVER['SCRIPT_NAME']
	 */
	public static function detectScriptPath () {
		// Default is empty
		$scriptPath = '';

		// Is the scriptname set?
		if (isset($_SERVER['SCRIPT_NAME'])) {
			// Get dirname from it and replace back-slashes with slashes for lame OSes...
			$scriptPath = str_replace("\\", '/', dirname($_SERVER['SCRIPT_NAME']));
		} // END - if

		// Return it
		return $scriptPath;
	}

	/**
	 * 1) Loads class scanner and scans all framework's classes and interfaces.
	 * This method also registers the class loader's method autoLoad() for the
	 * SPL auto-load feature. Remember that you can register additional methods
	 * (not functions, please) for other libraries.
	 *
	 * Yes, I know about Composer, but I like to keep my class loader around.
	 * You can always use mine as long as your classes have a namespace
	 * according naming-convention: Vendor\Project\Group[\SubGroup]
	 *
	 * @return	void
	 */
	private static function scanFrameworkClasses () {
		// Include the class loader function
		require self::getConfigurationInstance()->getConfigEntry('framework_base_path') . 'loader/class_ClassLoader.php';

		// Register auto-load function with the SPL
		spl_autoload_register('Org\Mxchange\CoreFramework\Loader\ClassLoader::autoLoad');

		// Scan for all framework classes, exceptions and interfaces
		ClassLoader::scanFrameworkClasses();
	}

	/**
	 * 2) Determines request/response type and stores the created
	 * request/response instances in this object for later usage.
	 *
	 * @return	void
	 */
	private static function determineRequestType () {
		// Determine request type
		$request = self::getRequestTypeFromSystem();
		$requestType = self::getRequestTypeFromSystem();

		// Create a new request object
		$requestInstance = ObjectFactory::createObjectByName(sprintf('Org\Mxchange\CoreFramework\Request\%sRequest', StringUtils::convertToClassName($request)));

		// Remember request instance here
		self::setRequestInstance($requestInstance);

		// Do we have another response?
		if ($requestInstance->isRequestElementSet('request')) {
			// Then use it
			$request = strtolower($requestInstance->getRequestElement('request'));
			$requestType = $request;
		} // END - if

		// ... and a new response object
		$responseClass = sprintf('Org\Mxchange\CoreFramework\Response\%sResponse', StringUtils::convertToClassName($request));
		$responseInstance = ObjectFactory::createObjectByName($responseClass);

		// Remember response instance here
		self::setResponseInstance($responseInstance);
	}

	/**
	 * 3) Validate parameter 'application' if it is set and the application is there.
	 *
	 * @return	void
	 */
	private static function validateApplicationParameter () {
		// Is the parameter set?
		if (!self::getRequestInstance()->isRequestElementSet('app')) {
			/*
			 * Don't continue here, the application 'selector' is no longer
			 * supported and only existed as an idea to select the proper
			 * application (by user).
			 */
			ApplicationEntryPoint::exitApplication('No application specified. Please provide a parameter "app" and retry.');
		} // END - if

		// Get it for local usage
		$application = self::getRequestInstance()->getRequestElement('app');

		// Secure it, by keeping out tags
		$application = htmlentities(strip_tags($application), ENT_QUOTES);

		// Secure it a little more with a reg.exp.
		$application = preg_replace('/([^a-z0-9_-])+/i', '', $application);

		// Construct FQPN (Full-Qualified Path Name) for ApplicationHelper class
		$applicationPath = sprintf(
			'%s%s%s',
			self::getConfigurationInstance()->getConfigEntry('application_base_path'),
			$application,
			DIRECTORY_SEPARATOR
		);

		// Full path for application
		// Is the path there? This secures a bit the parameter (from untrusted source).
		if ((!is_dir($applicationPath)) || (!is_readable($applicationPath))) {
			// Not found or not readable
			ApplicationEntryPoint::exitApplication(sprintf('Application "%s" not found.', $application));
		} // END - if

		// Set the detected application's name and full path for later usage
		self::getConfigurationInstance()->setConfigEntry('detected_full_app_path', $applicationPath);
		self::getConfigurationInstance()->setConfigEntry('detected_app_name'     , $application);
	}

	/**
	 * Getter for request instance
	 *
	 * @return	$requestInstance	An instance of a Requestable class
	 */
	public static function getRequestInstance () {
		return self::$requestInstance;
	}

	/**
	 * Getter for response instance
	 *
	 * @return	$responseInstance	An instance of a Responseable class
	 */
	public static function getResponseInstance () {
		return self::$responseInstance;
	}

	/**
	 * Setter for request instance
	 *
	 * @param	$requestInstance	An instance of a Requestable class
	 * @return	void
	 */
	private static function setRequestInstance (Requestable $requestInstance) {
		self::$requestInstance = $requestInstance;
	}

	/**
	 * Setter for response instance
	 *
	 * @param	$responseInstance	An instance of a Responseable class
	 * @return	void
	 */
	private static function setResponseInstance (Responseable $responseInstance) {
		self::$responseInstance = $responseInstance;
	}

	/**
	 * Setter for database instance
	 *
	 * @param	$databaseInstance	An instance of a DatabaseConnection class
	 * @return	void
	 */
	public static function setDatabaseInstance (DatabaseConnection $databaseInstance) {
		self::$databaseInstance = $databaseInstance;
	}

	/**
	 * Getter for database instance
	 *
	 * @return	$databaseInstance	An instance of a DatabaseConnection class
	 */
	public static function getDatabaseInstance () {
		// Return instance
		return self::$databaseInstance;
	}

	/**
	 * Private getter for language instance
	 *
	 * @return	$languageInstance	An instance of a ManageableLanguage class
	 */
	public static function getLanguageInstance () {
		return self::$languageInstance;
	}

	/**
	 * Setter for language instance
	 *
	 * @param	$languageInstance	An instance of a ManageableLanguage class
	 * @return	void
	 */
	public static function setLanguageInstance (ManageableLanguage $languageInstance) {
		self::$languageInstance = $languageInstance;
	}

}
