<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Filesytem\Pointer;

// Import framework stuff
use Org\Mxchange\CoreFramework\Filesystem\FrameworkDirectory;
use Org\Mxchange\CoreFramework\Generic\NullPointerException;
use Org\Mxchange\CoreFramework\Object\BaseFrameworkSystem;

// Deprecated imports
use Org\Mxchange\CoreFramework\Deprecated\PathIsNoDirectoryException;

// Import SPL stuff
use \DirectoryIterator;

/**
 * A class for directory reading and getting its contents, no recursion!
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2020 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class FrameworkDirectoryPointer extends BaseFrameworkSystem implements FrameworkDirectory {
	/**
	 * The current path we are working in
	 */
	private $pathName = '';

	/**
	 * The directory iterator instance
	 */
	private $iteratorInstance = NULL;

	/**
	 * Protected constructor
	 */
	protected function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Destructor for cleaning purposes, etc
	 */
	public function __destruct() {
		// Is there a resource pointer? Then we have to close the directory here!
		if ($this->getDirectoryIteratorInstance() instanceof DirectoryIterator) {
			// Try to close a directory
			$this->closeDirectory();
		} // END - if

		// Call the parent destructor
		parent::__destruct();
	}

	/**
	 * Create a directory pointer based on the given path. The path will also
	 * be verified here.
	 *
	 * @param	$pathName		The path name we shall pass to opendir()
	 * @return	$pointerInstance	A prepared instance of FrameworkDirectoryPointer
	 * @throws	NullPointerException	If the provided path name is null
	 * @throws	InvalidPathStringException	If the provided path name is not a string
	 * @throws	PathIsNoDirectoryException	If the provided path name is not valid
	 * @throws	PathReadProtectedException	If the provided path name is read-protected
	 */
	public static final function createFrameworkDirectoryPointer ($pathName) {
		// Some pre-sanity checks...
		if (is_null($pathName)) {
			// No pathname given
			throw new NullPointerException($this, self::EXCEPTION_IS_NULL_POINTER);
		} elseif (!is_string($pathName)) {
			// Is not a string
			throw new InvalidPathStringException(NULL, self::EXCEPTION_INVALID_STRING);
		} elseif (!is_dir($pathName)) {
			// Not a directory
			throw new PathIsNoDirectoryException($pathName, self::EXCEPTION_INVALID_PATH_NAME);
		} elseif (!is_readable($pathName)) {
			// Not readable
			throw new PathReadProtectedException($pathName, self::EXCEPTION_READ_PROTECED_PATH);
		}

		// Create new instance
		$pointerInstance = new FrameworkDirectoryPointer();

		// Get an iterator for the directory
		$iteratorInstance = new DirectoryIterator($pathName);

		// ... and rewind back
		$iteratorInstance->rewind();

		// Set directory pointer and path name
		$pointerInstance->setDirectoryIteratorInstance($iteratorInstance);
		$pointerInstance->setPathName($pathName);

		// Return the instance
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('DIRECTORY-POINTER: Opened pathName=' . $pathName . ' - EXIT!');
		return $pointerInstance;
	}

	/**
	 * Read raw lines of data from a directory pointer and return the data
	 *
	 * @return	$currentEntry	Current entry from encapsulated iterator
	 */
	public function readRawDirectory () {
		// Can the next entry be read?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('DIRECTORY-POINTER: CALLED!');
		assert($this->getDirectoryIteratorInstance()->valid());

		// Read data from the directory pointer and return it
		$currentEntry = $this->getDirectoryIteratorInstance()->current();

		// Return found entry
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('DIRECTORY-POINTER: currentEntry[%s]=%s - EXIT!', gettype($currentEntry), $currentEntry));
		return $currentEntry;
	}

	/**
	 * Read lines from the current directory pointer except some parts
	 *
	 * @param	$except		Some parts of a directory we want to ignore. Valid: directory and file names, other values will be silently ignored
	 * @return	SplFileInfo		An instance of a SplFileInfo class
	 */
	public function readDirectoryExcept (array $except = []) {
		// No exceptions given?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('DIRECTORY-POINTER: except()=%d - CALLED!', count($except)));
		if (count($except) == 0) {
			// No exception given, so read all files and directories, but not recursive
			self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('DIRECTORY-POINTER: No exceptions given, please use readRawDirectory() instead!');
			return $this->readRawDirectory();
		} elseif (!$this->getDirectoryIteratorInstance()->valid()) {
			// No more left to read
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('DIRECTORY-POINTER: EOD reached - EXIT!');
			return NULL;
		}

		// Init raw line
		$fileInfoInstance = NULL;

		// Read a raw line...
		$currentEntry = $this->readRawDirectory();

		// Shall we exclude directories?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('DIRECTORY-POINTER: currentEntry[]=%s', gettype($currentEntry)));
		if (is_object($currentEntry)) {
			// Get file name
			$fileInfoInstance = $currentEntry;

			// Is it a dot-directory or excluded?
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('DIRECTORY-POINTER: fileInfoInstance->filename=%s,isDot=%d', $fileInfoInstance->getFilename(), intval($this->getDirectoryIteratorInstance()->isDot())));
			while ($this->getDirectoryIteratorInstance()->valid() && ($this->getDirectoryIteratorInstance()->isDot() || in_array($fileInfoInstance->getFilename(), $except))) {
				// Update current instance
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('DIRECTORY-POINTER: fileInfoInstance->filename=%s,isDot=%d', $fileInfoInstance->getFilename(), intval($this->getDirectoryIteratorInstance()->isDot())));
				$fileInfoInstance = $this->readRawDirectory();

				// To next entry
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('DIRECTORY-POINTER: Invoking directoryIteratorInstance->next() ...');
				$this->getDirectoryIteratorInstance()->next();
			} // END - while
		} // END - if

		// Return read line
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('DIRECTORY-POINTER: fileInfoInstance[%s]=%s - EXIT!', gettype($fileInfoInstance), $fileInfoInstance));
		return $fileInfoInstance;
	}

	/**
	 * Close a directory source and set it's instance to null and the path name
	 * to empty
	 *
	 * @return	void
	 */
	public function closeDirectory () {
		// Close the directory by unsetting it
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('DIRECTORY-POINTER: CALLED!');
		$this->unsetDirectoryIteratorInstance();
		$this->setPathName('');
	}

	/**
	 * Setter for the directory pointer
	 *
	 * @param	$iteratorInstance	An instanceof a DirectoryIterator class
	 * @return	void
	 */
	protected final function setDirectoryIteratorInstance (DirectoryIterator $iteratorInstance) {
		// Set instance
		$this->iteratorInstance = $iteratorInstance;
	}

	/**
	 * Getter for the directory pointer
	 *
	 * @return	$iteratorInstance	The directory pointer which shall be a valid directory resource
	 */
	public final function getDirectoryIteratorInstance () {
		return $this->iteratorInstance;
	}

	/**
	 * Remove directory iterator instance (effectively closing it) by setting
	 * it to NULL. This will trigger a call on the destructor which will then
	 * "close" the iterator.
	 *
	 * @param	$iteratorInstance	An instanceof a DirectoryIterator class
	 * @return	void
	 */
	protected final function unsetDirectoryIteratorInstance () {
		// "Unset" the instance
		$this->iteratorInstance = NULL;
	}

	/**
	 * Setter for path name
	 *
	 * @param	$pathName	The new path name
	 * @return	void
	 */
	protected final function setPathName ($pathName) {
		$pathName = (string) $pathName;
		$this->pathName = $pathName;
	}

	/**
	 * Getter for path name
	 *
	 * @return	$pathName	The current path name
	 */
	public final function getPathName () {
		return $this->pathName;
	}

}
